import pytest

from beanie import After, Before
from tests.odm.models import (
    DocumentWithActions,
    DocumentWithActions2,
    DocumentWithActionWinsStrategy,
    DocumentWithUnderscoreAction,
    DocumentWithUpdateFieldAction,
    DocumentWithValidateOnSaveAction,
    InheritedDocumentWithActions,
)


class TestActions:
    @pytest.mark.parametrize(
        "doc_class",
        [
            DocumentWithActions,
            DocumentWithActions2,
            InheritedDocumentWithActions,
        ],
    )
    async def test_actions_insert(self, doc_class):
        test_name = f"test_actions_insert_{doc_class}"
        sample = doc_class(name=test_name)

        await sample.insert()
        assert sample.name != test_name
        assert sample.name == test_name.capitalize()
        assert sample.num_1 == 1
        assert sample.num_2 == 9

    async def test_underscore_prefixed_action(self):
        sample = DocumentWithUnderscoreAction(name="test")

        await sample.insert()

        assert sample.creator_id == "some_user"

        refetched = await DocumentWithUnderscoreAction.find_one(
            {"_id": sample.id}
        )
        assert refetched.creator_id == "some_user"

    @pytest.mark.parametrize(
        "doc_class",
        [
            DocumentWithActions,
            DocumentWithActions2,
            InheritedDocumentWithActions,
        ],
    )
    async def test_actions_replace(self, doc_class):
        test_name = f"test_actions_replace_{doc_class}"
        sample = doc_class(name=test_name)

        await sample.insert()

        await sample.replace()
        assert sample.num_1 == 2
        assert sample.num_3 == 99

    @pytest.mark.parametrize(
        "doc_class",
        [
            DocumentWithActions,
            DocumentWithActions2,
            InheritedDocumentWithActions,
        ],
    )
    async def test_skip_actions_insert(self, doc_class):
        test_name = f"test_skip_actions_insert_{doc_class}"
        sample = doc_class(name=test_name)

        await sample.insert(skip_actions=[After, "capitalize_name"])
        # capitalize_name has been skipped
        assert sample.name == test_name
        # add_one has not been skipped
        assert sample.num_1 == 1
        # num_2_change has been skipped
        assert sample.num_2 == 10

    @pytest.mark.parametrize(
        "doc_class",
        [
            DocumentWithActions,
            DocumentWithActions2,
            InheritedDocumentWithActions,
        ],
    )
    async def test_skip_actions_replace(self, doc_class):
        test_name = f"test_skip_actions_replace{doc_class}"
        sample = doc_class(name=test_name)

        await sample.insert()

        await sample.replace(skip_actions=[Before, "num_3_change"])
        # add_one has been skipped
        assert sample.num_1 == 1
        # num_3_change has been skipped
        assert sample.num_3 == 100

    @pytest.mark.parametrize(
        "doc_class",
        [
            DocumentWithActions,
            DocumentWithActions2,
            InheritedDocumentWithActions,
        ],
    )
    async def test_actions_delete(self, doc_class):
        test_name = f"test_actions_delete_{doc_class}"
        sample = doc_class(name=test_name)

        await sample.delete()
        assert sample.Inner.inner_num_1 == 1
        assert sample.Inner.inner_num_2 == 2

    @pytest.mark.parametrize(
        "doc_class",
        [
            DocumentWithActions,
            DocumentWithActions2,
            InheritedDocumentWithActions,
        ],
    )
    async def test_actions_update(self, doc_class):
        test_name = f"test_actions_update_{doc_class}"
        sample = doc_class(name=test_name)
        await sample.insert()

        await sample.update({"$set": {"name": "new_name"}})
        assert sample.name == "new_name"
        assert sample.num_1 == 1
        assert sample.num_2 == 9
        assert sample._private_num == 101

        await sample.set({"name": "awesome_name"})

        assert sample._private_num == 102
        assert sample.num_2 == 9
        assert sample.name == "awesome_name"

    @pytest.mark.parametrize(
        "doc_class",
        [
            DocumentWithActions,
            DocumentWithActions2,
            InheritedDocumentWithActions,
        ],
    )
    async def test_actions_save(self, doc_class):
        test_name = f"test_actions_save_{doc_class}"
        sample = doc_class(name=test_name)
        await sample.save()
        assert sample.num_1 == 1


class TestBeforeEventUpdatePersistence:
    """Tests for #1103: before_event(Update) changes must be persisted to DB."""

    async def test_update_persists_before_event_changes(self):
        doc = DocumentWithUpdateFieldAction(name="alice")
        await doc.insert()

        await doc.update({"$set": {"name": "bob"}})
        assert doc.name == "bob"
        assert doc.tag == "updated"

        refetched = await DocumentWithUpdateFieldAction.find_one(
            {"_id": doc.id}
        )
        assert refetched.tag == "updated"
        assert refetched.name == "bob"

    async def test_set_persists_before_event_changes(self):
        doc = DocumentWithUpdateFieldAction(name="alice")
        await doc.insert()

        await doc.set({"name": "charlie"})
        assert doc.tag == "updated"

        refetched = await DocumentWithUpdateFieldAction.find_one(
            {"_id": doc.id}
        )
        assert refetched.tag == "updated"
        assert refetched.name == "charlie"

    async def test_update_skip_actions_before(self):
        doc = DocumentWithUpdateFieldAction(name="alice")
        await doc.insert()

        await doc.update({"$set": {"name": "bob"}}, skip_actions=[Before])

        refetched = await DocumentWithUpdateFieldAction.find_one(
            {"_id": doc.id}
        )
        assert refetched.tag is None
        assert refetched.name == "bob"


class TestValidateOnSaveWithBeforeEvent:
    """Tests for #894: before_event must run after validate_on_save."""

    async def test_insert_before_event_survives_validation(self):
        doc = DocumentWithValidateOnSaveAction(name="  Hello World  ")
        await doc.insert()

        assert doc.normalized_name == "hello world"

        refetched = await DocumentWithValidateOnSaveAction.find_one(
            {"_id": doc.id}
        )
        assert refetched.normalized_name == "hello world"

    async def test_save_before_event_survives_validation(self):
        doc = DocumentWithValidateOnSaveAction(name="initial")
        await doc.insert()

        doc.name = "  Updated Name  "
        await doc.save()

        assert doc.normalized_name == "updated name"

        refetched = await DocumentWithValidateOnSaveAction.find_one(
            {"_id": doc.id}
        )
        assert refetched.normalized_name == "updated name"

    async def test_replace_before_event_survives_validation(self):
        doc = DocumentWithValidateOnSaveAction(name="initial")
        await doc.insert()

        doc.name = "  Replaced Name  "
        await doc.replace()

        assert doc.normalized_name == "replaced name"

        refetched = await DocumentWithValidateOnSaveAction.find_one(
            {"_id": doc.id}
        )
        assert refetched.normalized_name == "replaced name"


class TestActionConflictResolutionIntegration:
    """Integration tests for action_conflict_resolution setting with real DB."""

    async def test_action_wins_overrides_explicit_set(self):
        doc = DocumentWithActionWinsStrategy(name="alice")
        await doc.insert()

        await doc.update({"$set": {"name": "bob", "tag": "explicit"}})

        refetched = await DocumentWithActionWinsStrategy.find_one(
            {"_id": doc.id}
        )
        assert refetched.name == "bob"
        assert refetched.tag == "action_value"

    async def test_action_wins_via_set_method(self):
        doc = DocumentWithActionWinsStrategy(name="alice")
        await doc.insert()

        await doc.set({"tag": "explicit"})

        refetched = await DocumentWithActionWinsStrategy.find_one(
            {"_id": doc.id}
        )
        assert refetched.tag == "action_value"

    async def test_update_wins_non_conflicting_fields_persisted(self):
        doc = DocumentWithUpdateFieldAction(name="alice")
        await doc.insert()

        await doc.update({"$set": {"name": "bob"}})

        refetched = await DocumentWithUpdateFieldAction.find_one(
            {"_id": doc.id}
        )
        assert refetched.name == "bob"
        assert refetched.tag == "updated"



import inspect
from typing import Any, get_args, get_origin

from beanie.odm.fields import IndexedAnnotation

from .pydantic import get_field_type


def extract_id_class(annotation) -> type[Any]:
    if get_origin(annotation) is not None:
        try:
            annotation = next(
                arg for arg in get_args(annotation) if arg is not type(None)
            )
        except StopIteration:
            annotation = None
    if inspect.isclass(annotation):
        return annotation
    raise ValueError(f"Unknown annotation: {annotation}")


def get_index_attributes(field) -> tuple[int, dict[str, Any]] | None:
    """Gets the index attributes from the field, if it is indexed.

    :param field: The field to get the index attributes from.

    :return: The index attributes, if the field is indexed. Otherwise, None.
    """
    # For fields that are directly typed with `Indexed()`, the type will have
    # an `_indexed` attribute.
    field_type = get_field_type(field)
    if hasattr(field_type, "_indexed"):
        return getattr(field_type, "_indexed", None)

    # For fields that are use `Indexed` within `Annotated`, the field will have
    # metadata that might contain an `IndexedAnnotation` instance.
    # In Pydantic 2, the field has a `metadata` attribute with
    # the annotations.
    metadata = getattr(field, "metadata", None)

    if metadata is None:
        return None

    try:
        iter(metadata)
    except TypeError:
        return None

    indexed_annotation = next(
        (
            annotation
            for annotation in metadata
            if isinstance(annotation, IndexedAnnotation)
        ),
        None,
    )

    return getattr(indexed_annotation, "_indexed", None)


def is_generic_alias(obj: Any) -> bool:
    """Check if the object is a typing or built-in generic alias (e.g., list[str], List[int]).

    :param obj: An object instance.

    :return: True if obj is a typing or built-in generic alias, False otherwise.
    """

    # Check built-in generic aliases (e.g. list[str])
    try:
        from types import GenericAlias

        if isinstance(obj, GenericAlias):
            return True
    except ImportError:
        pass  # Python < 3.9

    # Check legacy typing generics (e.g. typing.List[str])
    try:
        from typing import _GenericAlias  # type: ignore[attr-defined]

        if isinstance(obj, _GenericAlias):
            return True
    except ImportError:
        pass

    return False



from pymongo.asynchronous.client_session import AsyncClientSession


class SessionMethods:
    """
    Session methods
    """

    def set_session(self, session: AsyncClientSession | None = None):
        """
        Set session
        :param session: Optional[AsyncClientSession] - pymongo session
        :return:
        """
        if session is not None:
            self.session: AsyncClientSession | None = session
        return self



from abc import ABC, abstractmethod

from beanie.odm.documents import Document


class BaseMigrationController(ABC):
    def __init__(self, function):
        self.function = function

    @abstractmethod
    async def run(self, session):
        pass

    @property
    @abstractmethod
    def models(self) -> list[type[Document]]:
        pass



from beanie.odm.queries.update import UpdateOne
from tests.typing.models import ProjectionTest, Test


async def find_many() -> list[Test]:
    return await Test.find().to_list()


async def find_many_with_projection() -> list[ProjectionTest]:
    return await Test.find().project(projection_model=ProjectionTest).to_list()


async def find_many_generator() -> list[Test]:
    docs: list[Test] = []
    async for doc in Test.find():
        docs.append(doc)
    return docs


async def find_many_generator_with_projection() -> list[ProjectionTest]:
    docs: list[ProjectionTest] = []
    async for doc in Test.find().project(projection_model=ProjectionTest):
        docs.append(doc)
    return docs


async def find_one() -> Test | None:
    return await Test.find_one()


async def find_one_with_projection() -> ProjectionTest | None:
    return await Test.find_one().project(projection_model=ProjectionTest)


def find_one_update() -> UpdateOne:
    return Test.find_one().update({"$set": {"bar": "updated"}})


def find_one_upsert() -> UpdateOne:
    return Test.find_one().upsert(
        {"$set": {"bar": "updated"}},
        on_insert=Test(foo="foo", bar="bar", baz="baz"),
    )



from functools import wraps
from typing import TYPE_CHECKING, TypeVar, cast

from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from beanie import Document
    from beanie.odm.documents import AsyncDocMethod

P = ParamSpec("P")
R = TypeVar("R")


def validate_self_before(
    f: "AsyncDocMethod[P, R]",
) -> "AsyncDocMethod[P, R]":
    @wraps(f)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self = cast("Document", args[0])
        await self.validate_self(*args, **kwargs)
        return await f(*args, **kwargs)

    return wrapper



from tests.odm.models import DocumentTestModel
from tests.odm.views import ViewForTest, ViewForTestWithLink


class TestViews:
    async def test_simple(self, documents):
        await documents(number=15)
        results = await ViewForTest.all().to_list()
        assert len(results) == 6

    async def test_aggregate(self, documents):
        await documents(number=15)
        results = await ViewForTest.aggregate(
            [
                {"$set": {"test_field": 1}},
                {"$match": {"$expr": {"$lt": ["$number", 12]}}},
            ]
        ).to_list()
        assert len(results) == 3
        assert results[0]["test_field"] == 1

    async def test_link(self, documents_with_links):
        await documents_with_links()
        results = await ViewForTestWithLink.all().to_list()
        for document in results:
            await document.fetch_all_links()

        for i, document in enumerate(results):
            assert document.link.test_int == i

    async def test_find_with_fetch_links(self, documents_with_links):
        """fetch_links=True should work with View.find() (#1011)."""
        await documents_with_links()
        results = await ViewForTestWithLink.find(
            fetch_links=True,
        ).to_list()
        assert len(results) > 0
        for document in results:
            assert isinstance(document.link, DocumentTestModel)
            assert document.link.test_int is not None

    async def test_find_one_with_fetch_links(self, documents_with_links):
        """fetch_links=True should work with View.find_one() (#1011)."""
        await documents_with_links()
        document = await ViewForTestWithLink.find_one(fetch_links=True)
        assert document is not None
        assert isinstance(document.link, DocumentTestModel)
        assert document.link.test_int is not None



import pytest
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from tests.fastapi.app import app


@pytest.fixture
async def api_client():
    """api client fixture."""
    async with LifespanManager(app, startup_timeout=100, shutdown_timeout=100):
        server_name = "http://localhost"
        async with AsyncClient(
            transport=ASGITransport(app=app), base_url=server_name
        ) as ac:
            yield ac



import inspect
from functools import wraps
from typing import TYPE_CHECKING, Any, TypeVar, cast, overload

from typing_extensions import ParamSpec

from beanie.exceptions import StateManagementIsTurnedOff, StateNotSaved

if TYPE_CHECKING:
    from beanie import Document
    from beanie.odm.documents import AnyDocMethod, AsyncDocMethod, DocType

P = ParamSpec("P")
R = TypeVar("R")


def check_if_state_saved(self: "DocType"):
    if not self.use_state_management():
        raise StateManagementIsTurnedOff(
            "State management is turned off for this document"
        )
    if self._saved_state is None:
        raise StateNotSaved("No state was saved")


@overload
def saved_state_needed(
    f: "AsyncDocMethod[P, R]",
) -> "AsyncDocMethod[P, R]": ...
@overload
def saved_state_needed(
    f: "AnyDocMethod[P, R]",
) -> "AnyDocMethod[P, R]": ...


def saved_state_needed(
    f: "AnyDocMethod[P, Any]",
) -> "AnyDocMethod[P, Any]":
    @wraps(f)
    def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        self = cast("Document", args[0])
        check_if_state_saved(self)
        return f(*args, **kwargs)

    @wraps(f)
    async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        self = cast("Document", args[0])
        check_if_state_saved(self)
        return await f(*args, **kwargs)

    if inspect.iscoroutinefunction(f):
        return async_wrapper
    return sync_wrapper


def check_if_previous_state_saved(self: "DocType"):
    if not self.use_state_management():
        raise StateManagementIsTurnedOff(
            "State management is turned off for this document"
        )
    if not self.state_management_save_previous():
        raise StateManagementIsTurnedOff(
            "State management's option to save previous state is turned off for this document"
        )


@overload
def previous_saved_state_needed(
    f: "AsyncDocMethod[P, R]",
) -> "AsyncDocMethod[P, R]": ...
@overload
def previous_saved_state_needed(
    f: "AnyDocMethod[P, R]",
) -> "AnyDocMethod[P, R]": ...


def previous_saved_state_needed(
    f: "AnyDocMethod[P, Any]",
) -> "AnyDocMethod[P, Any]":
    @wraps(f)
    def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        self = cast("Document", args[0])
        check_if_previous_state_saved(self)
        return f(*args, **kwargs)

    @wraps(f)
    async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        self = cast("Document", args[0])
        check_if_previous_state_saved(self)
        return await f(*args, **kwargs)

    if inspect.iscoroutinefunction(f):
        return async_wrapper
    return sync_wrapper


def save_state_after(
    f: "AsyncDocMethod[P, R]",
) -> "AsyncDocMethod[P, R]":
    @wraps(f)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self = cast("Document", args[0])
        result = await f(*args, **kwargs)
        self._save_state()
        return result

    return wrapper
